% Function pp2xyz
% Function pp2xyz creates a matrix in the form or {x y z} triplets.
% The program allows for the optional output of information to a 
% specified file.
% Usage:  [newmat] = pp2xyz (Zmat, grd_struct, (file_id, file_fmt) )
% Inputs:
%   Zmat - The  Z value matrix generated by functions ppzinit or ppsmooth.
%   grd_struct - the structure storing the grid description;
%      generated and returned by ppzinit or ppsmooth, it has the form.
%       grd_struct = struct ('x_min', XVEC(1), ...
%                     'x_max', XVEC(nx), ...
%                     'y_min', YVEC(1), ...
%                     'y_max', YVEC(ny), ...
%                     'x_inc', DX, ...
%                     'y_inc', DY, ...
%                     'nx', NX, ...
%                     'ny', NY, ...
%                     'missing', empty, ...
%                     'masked', mask_val);
%   file_id - optional file identifier for output.
%   file_fmt - optional format for output file.
%              default:  '%g %g %g\n'
% Outputs:
%   newmat - the generated matrix containing of {x y z} in columns.



function [newmat] = pp2xyz (Zmat, grd_struct, varargin)

newmat = [];
     % Check for arguments.
if (nargin <2)
  fprintf (2, '   *** pp2xyz ERROR:  NOT ENOUGH INPUT ARGUMENTS.')
  return;
  end  % if enough argments.

xmin = grd_struct.x_min;
xmax = grd_struct.x_max;
xdel = grd_struct.x_inc;
NX = grd_struct.nx;
ymin = grd_struct.y_min;
ymax = grd_struct.y_max;
ydel = grd_struct.y_inc;
NY = grd_struct.ny;

     % Compute nodes this way to make sure there are no round-off problems/
XVEC = xmin + (0:(NX-1))*xdel;
YVEC = ymin + (0:(NY-1))*ydel;

if (nargin >= 3)
  fid = fopen (varargin{1}, 'w');
  file_fmt = '%g %g %g\n';

     % Output the header.  This means you have to use the -H1 option 
     % on the invocation of the GMT routine xyz2grd.
  fprintf (fid, 'Header: %g %g  %g %g  %g %g  %d %d  %g %g\n', ...
           xmin, xmax,  ymin, ymax,  xdel, ydel,  NX, NY, ...
           grd_struct.missing, grd_struct.masked);

  if (nargin == 4)
    file_fmt = varargin{2};
  end  % if format specified.

  disp (['  Program pp2xyz writing to file: '  varargin{1}]);
  disp (['  Program pp2xyz using format: '''  file_fmt  '''']);
  end  % if file specified.

k = 0;
for (i = 1:NX)
  for (j = 1:NY)
    k = k + 1;
    newmat(k,1) = XVEC(i);
    newmat(k,2) = YVEC(j);
    newmat(k,3) = Zmat(j,i);
    if (nargin >= 3)
      fprintf (fid, file_fmt, newmat(k,:) );
      end  % if file.
    end  % for i
  end  % for j

if (nargin >= 3)  
  fclose (fid);  
  disp (['  pp2xyz created file: '  varargin{1}])
  end;

return 


